/*
** iplog_pcap.c - iplog pcap management routines.
** Copyright (C) 1999-2000 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**
** $Id: iplog_pcap.c,v 1.20 2000/02/18 03:35:37 odin Exp $
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <pcap.h>

#ifdef HAVE_SYS_SOCKIO_H
#	include <sys/sockio.h>
#endif

#include <iplog.h>
#include <iplog_pcap.h>
#include <iplog_options.h>

extern u_long flags;

static int get_if_namesaddr(struct pcap_data **int_list, const u_char **vif);
static int get_pcap_datalink(pcap_t *pd);

u_char *pcap_network = NULL;

/*
** XXX There's something broken with this.. some interfaces aren't
** being detected for people... I'm not sure if this is still the case or what
*/

static int get_if_namesaddr(struct pcap_data **int_list, const u_char **vif) {
	int sock;
#ifdef HAVE_SOCKADDR_SA_LEN
	u_int i;
#endif
	u_int z;
	struct ifconf ifc;
	struct ifreq buf[16], *ifr, *last, *next, ifrf;
	struct pcap_data data;
	struct sockaddr_in *gn_sin;
	u_char *p, *nlast = "";

	memset(&data, 0, sizeof(data));

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1)
		return (-1);

	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = (caddr_t) buf;

	if (ioctl(sock, SIOCGIFCONF, &ifc) != 0 ||
		(u_long) ifc.ifc_len < sizeof(struct ifreq))
	{
		close(sock);
		return (-1);
	}

	ifr = buf;
	last = (struct ifreq *) ((char *) buf + ifc.ifc_len);
	for (; ifr < last ; ifr = next) {
#ifdef HAVE_SOCKADDR_SA_LEN
		i = ifr->ifr_addr.sa_len + sizeof(ifr->ifr_name);
		if (i < sizeof(struct ifreq))
			next = ifr + 1;
		else
			next = (struct ifreq *) ((char *) ifr + i);
#else
		next = ifr + 1;
#endif
		xstrncpy(ifrf.ifr_name, ifr->ifr_name, sizeof(ifrf.ifr_name));

		if (ioctl(sock, SIOCGIFFLAGS, &ifrf) != 0)
			continue;
		if ((ifrf.ifr_flags & IFF_UP) && ifr->ifr_addr.sa_family == AF_INET) {
			if (vif != NULL) {
				bool found = false;
				u_char *a = xstrdup(ifr->ifr_name);
				u_char *b = strchr(a, ':');

				if (b != NULL)
					*b = '\0';

				for (z = 0 ; vif[z] != NULL ;) {
					if (strchr(vif[z], ':') != NULL)
						b = ifr->ifr_name;
					else
						b = a;

					if (!strcmp(vif[z++], b)) {
						found = true;
						break;
					}
				}

				free(a);
				if (found == false)
					continue;
			} else if (IS_LOOPBACK(ifr->ifr_name))
				continue;

			p = strchr(ifr->ifr_name, ':');
			if (p != NULL)
				*p = '\0';

			if (strcmp(nlast, ifr->ifr_name)) {
				if (strcmp(nlast, ""))
					list_copy_append(&data, int_list, sizeof(data));
				nlast = ifr->ifr_name;
				memset(&data, 0, sizeof(data));
				xstrncpy(data.name, ifr->ifr_name, sizeof(data.name));
			}
			gn_sin = (struct sockaddr_in *) &ifr->ifr_addr;
			data.addr[data.n_addr++] = gn_sin->sin_addr.s_addr;
		}
	}

	if (data.n_addr != 0)
		list_copy_append(&data, int_list, sizeof(data));

	close(sock);
	return (0);
}

static int get_pcap_datalink(pcap_t *pd) {
	int datalink = pcap_datalink(pd), dlt = -1;

	switch (datalink) {
		case DLT_RAW:
			dlt = 0;
			break;
		case DLT_ATM_RFC1483:
			dlt = 8;
			break;
		case DLT_EN10MB:
		case DLT_IEEE802:
			dlt = sizeof(struct ethhdr);
			break;
		case DLT_SLIP_BSDOS:
		case DLT_PPP_BSDOS:
			dlt = 24;
			break;
		case DLT_SLIP:
			dlt = 16;
			break;
		case DLT_PPP:
		case DLT_NULL:
#ifdef __OpenBSD__
		case DLT_LOOP:
#endif
			dlt = 4;
			break;
	}

	return (dlt);
}

bool interface_is_up(const u_char *name) {
	struct ifreq ifreq;
	int sock;
	bool ret = false;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1)
		return (false);

	if (strlen(name) >= sizeof(ifreq.ifr_name))
		goto out;

	xstrncpy(ifreq.ifr_name, name, sizeof(ifreq.ifr_name));

	if (ioctl(sock, SIOCGIFFLAGS, &ifreq) != 0)
		goto out;

	if (ifreq.ifr_flags & IFF_UP)
		ret = true;

out:
	close(sock);
	return (ret);
}

int open_pcap_device(struct pcap_data *pdata) {
	u_char fstring[1024], *temp, errbuf[2048];
	struct bpf_program filt;
	u_int i = 0;

	pdata->pd =
		pcap_open_live(pdata->name, 1550, !!(flags & PROMISC), 0, errbuf);

	if (pdata->pd == NULL)
		return (-1);

	pdata->dl = get_pcap_datalink(pdata->pd);

	if (pdata->dl == -1)
		return (-1);

#ifdef __linux__
	/*
	** pcap filters don't work on Linux loopback devices.
	** Each packet appears to be received twice.  I don't know why anyone
	** would want to monitor loopback, anyway.
	*/
	if (IS_LOOPBACK(pdata->name))
		return (0);
#endif

	xstrncpy(fstring, "ip and (", sizeof(fstring));
	if (flags & LOG_TCP) {
		xstrncat(fstring, "tcp ", sizeof(fstring));
		++i;
	}

	if (flags & LOG_UDP) {
		if (i++ == 0)
			temp = "udp ";
		else
			temp = "or udp ";
		xstrncat(fstring, temp, sizeof(fstring));
	}

	if (flags & LOG_ICMP) {
		if (i == 0)
			temp = "icmp";
		else
			temp = "or icmp";
		xstrncat(fstring, temp, sizeof(fstring));
	}

	if (flags & PROMISC) {
		u_char *p, *cur, *orig = xstrdup(pcap_network);

		cur = orig;
		xstrncat(fstring, ") and dst net (", sizeof(fstring));

		while (1) {
			p = strchr(cur, ',');
			if (p != NULL)
				*p++ = '\0';
			xstrncat(fstring, cur, sizeof(fstring));
			if (p == NULL)
				break;
			xstrncat(fstring, " or ", sizeof(fstring));
			cur = p;
		}

		free(orig);
	} else {
		struct sockaddr_in opd_sin;
		u_char nbuf[16];

		xstrncat(fstring, ") and dst host (", sizeof(fstring));

		for (i = 0 ; i < pdata->n_addr ; i++) {
			opd_sin.sin_addr.s_addr = pdata->addr[i];
			if (i > 0)
				xstrncat(fstring, " or ", sizeof(fstring));
			inet_ntoa_r(&opd_sin.sin_addr, nbuf, sizeof(nbuf));
			xstrncat(fstring, nbuf, sizeof(fstring));
		}
	}

	xstrncat(fstring, ")", sizeof(fstring));

	if (pcap_compile(pdata->pd, &filt, fstring, 1, 0) == -1)
		fatal("pcap_compile: %s", pcap_geterr(pdata->pd));

	if (pcap_setfilter(pdata->pd, &filt) == -1)
		fatal("pcap_setfilter: %s", pcap_geterr(pdata->pd));

	return (0);
}

int setup_pcap(struct pcap_data **plist, const u_char *ifstring) {
	struct pcap_data *cur;
	u_char *vif[MAX_ADDR];
	u_int interfaces = 0;

	if (ifstring != NULL) {
		u_int i = 0;
		u_char *delim, *pos;
		u_char *tstr = xstrdup(ifstring);

		for (pos = tstr ; i < ARRAY_SIZE(vif) - 1
			&& (delim = strchr(pos, ',')) != NULL ; pos = delim)
		{
			*delim++ = '\0';
			vif[i++] = pos;
		}
		vif[i++] = pos;
		vif[i] = NULL;
		get_if_namesaddr(plist, (const u_char **) vif);
		free(tstr);
	} else
		get_if_namesaddr(plist, NULL);

	cur = *plist;
	while (cur != NULL) {
		if (open_pcap_device(cur) == -1) {
			mysyslog("Warning: unable to open %s", cur->name);
			cur = list_delete(cur, plist);
		} else {
			interfaces++;
			cur = cur->next;
		}
	}

	if (interfaces == 0)
		return (-1);

	return (0);
}

/* vim:ts=4:sw=8:tw=0 */
