/*
** main.c - iplog main routine.
** Copyright (C) 1999-2000 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**
** $Id: main.c,v 1.25 2000/02/18 03:35:37 odin Exp $
*/

#define _GNU_SOURCE

#include <config.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <pthread.h>
#include <pcap.h>

#ifdef HAVE_PATHS_H
#	include <paths.h>
#endif

#ifdef HAVE_PTHREAD_CANCEL
#	include <setjmp.h>
#endif

#ifndef __linux__
#	include <pcap-int.h>
#endif

#include <iplog.h>
#include <iplog_options.h>
#include <iplog_filter.h>
#include <iplog_pcap.h>
#include <iplog_input.h>
#include <iplog_dns.h>
#include <iplog_scan.h>

#define EXPIRE_INTERVAL 1

#define ANY_SCAN \
(PORTSCAN | NULL_SCAN | FIN_SCAN | XMAS_SCAN | UDP_SCAN | PING_FLOOD | SMURF)

struct running {
	struct running *next;
	pthread_t pt;
};

/* There's no pthread_cancel() on FreeBSD and BSDI (others?) */
#ifdef HAVE_PTHREAD_CANCEL
static void iplog_restart(int sig);
static void pcap_cleanup(void *data);

static sigjmp_buf jmpbuf;
#endif

static struct running *running = NULL;
static struct pcap_data *plist = NULL;
static volatile bool cap_rawsock = false;
static u_int reading = 0;

static void *read_packets(void *data);
static void *expire_data(void *unused);
static void iplog_cleanup(int sig);

u_char *rules = "/etc/iplog.rules", *ifstring, *logfile, *user, *group;

u_long flags = ~0 &	~(	GET_IDENT | NO_RESOLV | FOOL_NMAP | NO_FORK |
						VERBOSE | LOG_STDOUT | IGNORE_NS | LOG_IP |
						PROMISC | LOG_DEST | SCANS_ONLY);

int main(int argc, char **argv) {
	struct pcap_data *cur;
	struct running r_temp;
	volatile bool dropped = false;
	int temp;

	memset(&r_temp, 0, sizeof(r_temp));

	get_options(argc, argv);
	myopenlog("iplog", LOG_PID | LOG_NDELAY);
	check_options();

	if (!(flags & NO_FORK))
		fork_to_back();

	write_lockfile();

	if (flags & DNS_CACHE)
		init_dns_table((flags & PROMISC) ? DNS_MAXSIZE_P : DNS_MAXSIZE_N);

	if (flags & ANY_SCAN)
		init_scan_table((flags & PROMISC) ? SCAN_TSIZE_P : SCAN_TSIZE_N);

	init_frag_table(FRAG_TSIZE);

#ifdef HAVE_PTHREAD_CANCEL
	sigsetjmp(jmpbuf, 1);
#endif

	if (plist == NULL && setup_pcap(&plist, ifstring) == -1)
		fatal("Couldn't initialize interfaces.");

	if (dropped == false) {
		drop_privs(user, group);
		dropped = true;
	}

	/* This is not the best check, but it's better than checking {e,}uid */
	temp = socket(AF_INET, SOCK_RAW, IPPROTO_TCP);
	if (temp != -1) {
		cap_rawsock = true;
		close(temp);
	}

	read_filters(rules);

	if (flags & IGNORE_NS)
		add_dns_ignore_rules();

	pthread_create(&r_temp.pt, NULL, expire_data, NULL);
	list_copy_prepend(&r_temp, &running, sizeof(r_temp));

	for (cur = plist ; cur != NULL ; cur = cur->next) {
		pthread_create(&r_temp.pt, NULL, read_packets, cur);
		list_copy_prepend(&r_temp, &running, sizeof(r_temp));
		reading++;
	}

#ifdef HAVE_PTHREAD_CANCEL
	signal(SIGHUP, iplog_restart);
#else
	signal(SIGHUP, SIG_IGN);
#endif

	if (flags & NO_FORK)
		signal(SIGINT, iplog_cleanup);
	signal(SIGTERM, iplog_cleanup);
	signal(SIGSEGV, iplog_cleanup);

	/* Sleeps forever */
	xsleep(0);

	/* NOTREACHED */
	exit(0);
}

static void *read_packets(void *data) {
	u_char *packet;
	struct pcap_pkthdr filler;
	struct pcap_data *pdev = (struct pcap_data *) data;
	struct running *cur;
	sigset_t iplog_sigset;
#ifndef __linux__
	struct timeval tv;
	int ret, fd = ((struct pcap *)pdev->pd)->fd;
	fd_set rfds;
#endif

#ifdef HAVE_PTHREAD_CANCEL
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
#endif

	sigemptyset(&iplog_sigset);
	if (flags & NO_FORK)
		sigaddset(&iplog_sigset, SIGINT);
	sigaddset(&iplog_sigset, SIGTERM);
	sigaddset(&iplog_sigset, SIGHUP);
	pthread_sigmask(SIG_BLOCK, &iplog_sigset, NULL);

#ifndef __linux__
	tv.tv_sec = 0;
	tv.tv_usec = 10000;

	FD_ZERO(&rfds);
#endif

	for (;;) {
#ifndef __linux__
		/* FIXME: This may not be needed on Solaris.. */

		FD_SET(fd, &rfds);
		/*
		** Cope with BSD read (brokenness?).  Is there some better way to do
		** this?  The descriptor only seems to be marked ready for reading
		** after it receives many packets, causing messages to be logged in
		** bursts..  After playing with this some more, this behavior
		** is only happening when programs that use libpcap are compiled with
		** -pthread on BSD.
		*/
		ret = select(fd + 1, &rfds, NULL, NULL, &tv);
		/*
		** Try to read anyway, even if select returns 0, as it doesn't seem
		** to work correctly to begin with when called on a bpf descriptor..
		*/
#endif
		packet = (u_char *) pcap_next(pdev->pd, &filler);
		if (packet != NULL) {
			packet += pdev->dl;
			parse_packet((struct ip *) packet);
		} else if (errno == ENETDOWN) {
			mysyslog("Warning: interface %s went down.", pdev->name);
			pcap_close(pdev->pd);

			if (cap_rawsock == false) {
				mysyslog("%s cannot be brought back up.", pdev->name);

				if (--reading == 0)
					fatal("No more interfaces open for reading.  Exiting.");

				cur = running;

				/* Can't ever hit NULL */
				while (cur->pt != pthread_self())
					cur = cur->next;

				list_delete(cur, &running);
				list_delete(pdev, &plist);

				pthread_exit(NULL);
			} else {
				for (;;) {
					if (open_pcap_device(pdev) != -1) {
						if (interface_is_up(pdev->name)) {
							mysyslog("%s was successfully reopened",
								pdev->name);
							break;
						} else
							pcap_close(pdev->pd);
					}

					/* Try again in 30 seconds. */
					xsleep(30);
				}
			}
		}
	}

	return (NULL);
}

static void *expire_data(void *unused) {
	sigset_t expire_sigset;

#ifdef HAVE_PTHREAD_CANCEL
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
#endif

	sigemptyset(&expire_sigset);

	if (flags & NO_FORK)
		sigaddset(&expire_sigset, SIGINT);
	sigaddset(&expire_sigset, SIGTERM);
	sigaddset(&expire_sigset, SIGHUP);
	pthread_sigmask(SIG_BLOCK, &expire_sigset, NULL);

	for (;;) {
		xsleep(EXPIRE_INTERVAL);

		expire_frags();

		if (flags & ANY_SCAN)
			expire_scans();

		if (flags & DNS_CACHE)
			expire_dns();
	}

	return (unused);
}

#ifdef HAVE_PTHREAD_CANCEL
static void pcap_cleanup(void *data) {
	struct pcap_data *cur = data;

	pcap_close(cur->pd);
}

static void iplog_restart(int sig) {
	struct running *cur;
	extern pthread_mutex_t log_lock, syn_lock, bogus_lock, udp_lock;

	/* whee */
	pthread_mutex_lock(&log_lock);
	pthread_mutex_lock(&syn_lock);
	pthread_mutex_lock(&bogus_lock);
	pthread_mutex_lock(&udp_lock);

	for (cur = running ; cur != NULL ;) {
		pthread_cancel(cur->pt);
		pthread_join(cur->pt, NULL);
		cur = list_delete(cur, &running);
	}

	reading = 0;

	if (flags & ANY_SCAN)
		destroy_scan_table();

	if (flags & DNS_CACHE)
		destroy_dns_cache();

	destroy_frag_table();

	if (cap_rawsock == true) {
		list_destroy(plist, pcap_cleanup);
		plist = NULL;
	}

	destroy_filter_list(FIL_TCP);
	destroy_filter_list(FIL_UDP);
	destroy_filter_list(FIL_ICMP);

	pthread_mutex_unlock(&log_lock);
	pthread_mutex_unlock(&syn_lock);
	pthread_mutex_unlock(&bogus_lock);
	pthread_mutex_unlock(&udp_lock);

	mysyslog("restarting iplog (pid %d)", getpid());

	signal(sig, iplog_restart);
	siglongjmp(jmpbuf, 1);
}
#endif

static void iplog_cleanup(int sig) {
	unlink(LOCKFILE);
	mysyslog("Caught signal %d, exiting", sig);
	exit(0);
}

/* vim:ts=4:sw=8:tw=0 */
